from dash import dcc, html
from dash.development.base_component import Component

from amora.models import Model


def layout(model: Model) -> Component:
    owner = model.owner()
    if owner:
        content = [
            "Owner: ",
            dcc.Link(model.owner(), href=f"/owners/{owner}"),
        ]
    else:
        content = ["Owner: Unowned"]

    return html.H6(children=content, id="model-data-owner")



import urllib.parse

import dash
import dash_bootstrap_components as dbc
from dash import dcc, html
from dash.development.base_component import Component

from amora.models import list_models_with_owner, owners_to_models_dict

dash.register_page(
    __name__,
    name="Data Owners",
    path_template="/owners/<owner>",
    path="/owners",
)


def list_models_owned_by(owner: str) -> Component:
    models = list(list_models_with_owner(owner=owner))
    if not models:
        return html.Div(f"There are no models owned by `{owner}`")

    return dbc.ListGroup(
        children=[
            dbc.ListGroupItem(
                children=[
                    dcc.Link(
                        model.unique_name(), href=f"/models/{model.unique_name()}"
                    ),
                    html.P(model.__model_config__.description),
                ]
            )
            for model, _file_path in models
        ]
    )


def model_owners_list() -> Component:
    return dbc.ListGroup(
        children=[
            dbc.ListGroupItem(
                dbc.Row(
                    children=[
                        dbc.Col(dcc.Link(owner, href=f"/owners/{owner}"), width=11),
                        dbc.Col(
                            dbc.Badge(
                                len(owned_models),
                                title=f"Owns {len(owned_models)} models",
                                color="primary",
                                className="me-1",
                            ),
                            width=1,
                        ),
                    ],
                    justify="between",
                )
            )
            for owner, owned_models in owners_to_models_dict().items()
        ]
    )


def layout(owner: str = None) -> Component:
    if owner:
        owner = urllib.parse.unquote(owner)
        content = [
            html.H1("Data Owners"),
            html.H4(owner),
            html.P("List of owned Data Models:"),
            list_models_owned_by(owner),
        ]
    else:
        content = [
            html.H1("Data Owners"),
            html.P("List of Data Owners and the associated Data Models:"),
            model_owners_list(),
        ]

    return dbc.Container(children=[dbc.Row(row) for row in content])



import hashlib
import inspect
from collections import UserDict
from functools import wraps
from typing import Callable, NamedTuple, Union

import pandas as pd
from sqlalchemy import MetaData, create_engine

from amora import logger
from amora.config import StorageCacheProviders, settings

local_engine = create_engine(
    f"sqlite:///{settings.LOCAL_ENGINE_SQLITE_FILE_PATH}",
    echo=settings.LOCAL_ENGINE_ECHO,
)
local_metadata = MetaData(schema=None)


class CacheKey(NamedTuple):
    """
    E.g:

    ```python
    CacheKey(
        func_module="amora.questions",
        func_name="answer_df",
        func_checksum="e6ccc38abffd7081822da108971e9d9c",
        suffix="how_many_data_points_where_acquired",
    )

    "amora.questions.answer_df.e6ccc38abffd7081822da108971e9d9c.how_many_data_points_where_acquired"
    ```

    """

    func_module: str
    func_name: str
    func_checksum: str
    suffix: str

    def __repr__(self):
        return str(self)

    def __str__(self):
        return ".".join(value for value in self if value)


class Cache(UserDict):
    """
    A `pandas.DataFrame` cache provider that uses the local File System or GCS
    """

    file_suffix: str = ".parquet"

    @property
    def type_(self) -> StorageCacheProviders:
        return settings.STORAGE_CACHE_PROVIDER

    def filepath_for_key(self, key: CacheKey) -> str:
        if self.type_ is StorageCacheProviders.gcs:
            blob_name = f"{key}{self.file_suffix}"
            gs_url = f"gs://{settings.STORAGE_GCS_BUCKET_NAME}/{blob_name}"
            return gs_url

        if self.type_ is StorageCacheProviders.local:
            blob_name = f"{key}{self.file_suffix}"
            return settings.STORAGE_LOCAL_CACHE_PATH.joinpath(blob_name).as_posix()

        raise NotImplementedError  # pragma: nocover

    @logger.log_execution()
    def __setitem__(self, key: CacheKey, value: pd.DataFrame):
        value.to_parquet(
            self.filepath_for_key(key), engine=settings.STORAGE_PARQUET_ENGINE
        )

    @logger.log_execution()
    def __getitem__(self, item: CacheKey) -> pd.DataFrame:
        try:
            return pd.read_parquet(self.filepath_for_key(item))
        except FileNotFoundError as e:
            raise KeyError from e


CACHE = Cache()
Cacheable = Callable[..., pd.DataFrame]


def cache(suffix: Union[Callable[..., str], None] = None):
    """
    Caches a `amora.storage.Cacheable` into the provider selected
    at `setting.STORAGE_CACHE_PROVIDER`.

    To disable the cache, set the env var `AMORA_STORAGE_CACHE_ENABLED` to false.

    ```python
    from datetime import datetime
    from time import sleep

    import pandas as pd

    from amora.storage import cache


    @cache()
    def a_slow_to_build_dataframe():
        sleep(3)
        return pd.DataFrame([{"a": 4, "b": 2}])


    t0 = datetime.now()
    print(a_slow_to_build_dataframe())
    print("Uncached execution:", datetime.now() - t0)

    t0 = datetime.now()
    print(a_slow_to_build_dataframe())
    print("Cached execution:", datetime.now() - t0)
    ```

    ```
       a  b
    0  4  2
    Uncached execution: 0:00:03.011570

       a  b
    0  4  2
    Cached execution: 0:00:00.005567
    ```

    If the cached function expects arguments, a `suffix` function should be provided
    to the `cache` decorator. The function will be called with the same arguments a the
    cached function, and a `str` should be returned.

    ```python
    @cache(suffix=lambda arg1, arg2: f"{arg1}_{arg2}")
    def cacheable_func(arg1, arg2):
        return pd.DataFrame([{"amora": arg1, "storage": arg2}])


    cacheable_func(4, 2)
    # CacheKey -> foo_module.cacheable_func.e6ccc38abffd7081822da108971e9d9c.4_2

    cacheable_func(5, 3)
    # CacheKey -> foo_module.cacheable_func.e6ccc38abffd7081822da108971e9d9c.5_3
    ```

    The key function can also be used to implement a cache TTL:

    ```python
    from datetime import date


    @cache(suffix=lambda: date.today().isoformat())
    def im_cached_for_one_day():
        sleep(3)
        return pd.DataFrame([[1, 2, 3], [4, 5, 6]])
    ```
    """

    def wrapper(fn: Cacheable):
        func_checksum = hashlib.md5(inspect.getsource(fn).encode("utf-8")).hexdigest()

        @wraps(fn)
        def decorator(*args, **kwargs):
            if not settings.STORAGE_CACHE_ENABLED:
                return fn(*args, **kwargs)

            cache_key = CacheKey(
                func_module=fn.__module__,
                func_name=fn.__name__,
                func_checksum=func_checksum,
                suffix=suffix(*args, **kwargs) if suffix else "",
            )

            try:
                return CACHE[cache_key]
            except KeyError:
                result = fn(*args, **kwargs)
                CACHE[cache_key] = result
                return result

        return decorator

    return wrapper



import hashlib
import json
from os.path import exists
from pathlib import Path
from typing import Dict, Optional, Set, Tuple

from _hashlib import HASH
from pydantic import BaseModel

from amora.config import settings
from amora.dag import DependencyDAG
from amora.models import Model, amora_model_from_name_list, list_models

BUF_SIZE = 65536


class ModelMetadata(BaseModel):
    stat: float
    size: float
    hash: str
    path: str
    deps: list


class Manifest(BaseModel):
    models: Dict[str, ModelMetadata]

    @classmethod
    def from_project(cls) -> "Manifest":
        dag = DependencyDAG.from_project()

        models_manifest: Dict[str, ModelMetadata] = {}

        for model, model_file_path in list_models():
            file_stats = model_file_path.stat()
            model_unique_name = model.unique_name()

            models_manifest[model_unique_name] = ModelMetadata(
                stat=file_stats.st_mtime,
                size=file_stats.st_size,
                hash=hash_file(model_file_path).hexdigest(),
                path=str(model_file_path),
                deps=[
                    dep for dep in dag.get_all_dependencies(source=model_unique_name)
                ],
            )

        return Manifest(models=models_manifest)

    @classmethod
    def load(cls) -> Optional["Manifest"]:
        try:
            with open(settings.manifest_path) as f:
                return Manifest(**(json.load(f)))
        except FileNotFoundError:
            return None

    def save(self) -> None:
        with open(settings.manifest_path, "w+") as f:
            json.dump(self.dict(), f, indent=2)

    def get_models_to_compile(
        self: "Manifest", previous_manifest: "Manifest"
    ) -> Set[Tuple[Model, Path]]:
        models_to_compile = set()
        deps_names_to_compile: Set = set()

        for model, model_file_path in list_models():
            model_unique_name = model.unique_name()

            model_current_manifest = self.models[model_unique_name]
            model_previous_manifest = previous_manifest.models.get(
                model_unique_name
            )  # model could not exist in previous

            compile_model = not model_previous_manifest or (
                model_current_manifest.size != model_previous_manifest.size
                or model_current_manifest.deps != model_previous_manifest.deps
                or (
                    model_current_manifest.stat > model_previous_manifest.stat
                    and (
                        not exists(model.target_path())
                        or model_current_manifest.hash != model_previous_manifest.hash
                    )
                )
            )

            if compile_model:
                models_to_compile.add((model, model_file_path))
                deps_names_to_compile = deps_names_to_compile.union(
                    model_current_manifest.deps
                )

        deps_to_compile = amora_model_from_name_list(deps_names_to_compile)
        models_to_compile = models_to_compile.union(deps_to_compile)

        return models_to_compile


def hash_file(file_path: Path) -> HASH:
    hash = hashlib.md5()
    with open(file_path, "rb") as f:
        while True:
            data = f.read(BUF_SIZE)
            if not data:
                break
            hash.update(data)
    return hash



from prometheus_flask_exporter.multiprocess import GunicornPrometheusMetrics

from amora.dash.config import settings


def when_ready(server):
    GunicornPrometheusMetrics.start_http_server_when_ready(settings.METRICS_PORT)


def child_exit(server, worker):
    GunicornPrometheusMetrics.mark_process_dead_on_child_exit(worker.pid)



def patch_online_store():
    """
    Forces Feast to update Redis TTL

    issue: https://github.com/feast-dev/feast/issues/3275
    """
    from functools import wraps

    from feast.infra.online_stores.helpers import _redis_key
    from feast.infra.online_stores.redis import RedisOnlineStore

    def expire_all(func):
        @wraps(func)
        def wrapper(self, config, table, data, progress):
            func(self, config, table, data, progress)

            online_store_config = config.online_store

            client = self._get_client(online_store_config)
            project = config.project

            with client.pipeline(transaction=False) as pipe:
                for entity_key, _, _, _ in data:
                    redis_key_bin = _redis_key(
                        project,
                        entity_key,
                        entity_key_serialization_version=config.entity_key_serialization_version,
                    )
                    if online_store_config.key_ttl_seconds:
                        pipe.expire(
                            name=redis_key_bin,
                            time=online_store_config.key_ttl_seconds,
                        )
                pipe.execute()

        return wrapper

    RedisOnlineStore.online_write_batch = expire_all(
        RedisOnlineStore.online_write_batch
    )



from dataclasses import dataclass
from datetime import datetime, timedelta
from pathlib import Path
from typing import Optional

import humanize
from google.api_core.exceptions import ClientError
from google.cloud.bigquery import (
    Client,
    PartitionRange,
    QueryJobConfig,
    RangePartitioning,
    Table,
    TimePartitioning,
)

from amora.models import (
    MaterializationTypes,
    Model,
    ModelConfig,
    amora_model_for_name,
    amora_model_for_target_path,
)
from amora.providers.bigquery import schema_for_model


@dataclass
class Task:
    sql_stmt: str
    model: Model
    target_file_path: Path

    @classmethod
    def for_target(cls, target_file_path: Path) -> "Task":
        return cls(
            sql_stmt=target_file_path.read_text(),
            model=amora_model_for_target_path(target_file_path),
            target_file_path=target_file_path,
        )

    def __repr__(self):
        return f"{self.model.unique_name()} -> {self.sql_stmt}"


@dataclass
class Result:
    model_name: str
    model_config: ModelConfig
    destination_table: Table
    total_bytes_billed: int = 0
    total_bytes_processed: int = 0
    duration: Optional[timedelta] = None

    def __str__(self):
        if self.model_config.materialized == MaterializationTypes.table:
            rows = humanize.intcomma(self.destination_table.num_rows)
            processed_bytes_ = humanize.naturalsize(self.total_bytes_processed)
            duration = humanize.naturaldelta(self.duration)
            table_bytes_ = humanize.naturalsize(self.destination_table.num_bytes)

            return f"[{self.model_name}] Took {duration} to process {processed_bytes_} of data and materialize it into a `Table` with {rows} rows and {table_bytes_}."
        elif self.model_config.materialized == MaterializationTypes.view:
            return f"[{self.model_name}] Materialized as `View`"
        else:
            raise ValueError


def materialize(sql: str, model_name: str, config: ModelConfig) -> Optional[Result]:
    materialization = config.materialized

    if materialization == MaterializationTypes.ephemeral:
        return None

    client = Client()
    client.delete_table(model_name, not_found_ok=True)
    model = amora_model_for_name(model_name)

    if materialization == MaterializationTypes.view:
        view = Table(model_name)
        view.description = config.description
        view.labels = config.labels_dict
        view.view_query = sql

        table = client.create_table(view)
        return Result(
            model_name=model_name, model_config=config, destination_table=table
        )

    if materialization == MaterializationTypes.table:
        table = Table(model_name, schema=schema_for_model(model))
        table.description = config.description
        table.labels = config.labels_dict
        table.clustering_fields = config.cluster_by

        if config.partition_by:
            if config.partition_by.data_type == "int":
                table.range_partitioning = RangePartitioning(
                    range_=PartitionRange(
                        start=config.partition_by.range.get("start"),
                        end=config.partition_by.range.get("end"),
                    ),
                    field=config.partition_by.field,
                )

            else:
                table.time_partitioning = TimePartitioning(
                    field=config.partition_by.field,
                    type_=config.partition_by.granularity.upper(),
                )

        if config.hours_to_expire:
            table.expires = datetime.utcnow() + timedelta(hours=config.hours_to_expire)

        client.create_table(table)
        try:
            query_job = client.query(
                sql,
                job_config=QueryJobConfig(destination=table),
            )
            _result = query_job.result()
        except ClientError as e:
            raise ValueError(
                f"Materialization failed for model `{model_name}` to destination `{table}`"
            ) from e
        else:
            return Result(
                model_name=model_name,
                model_config=config,
                destination_table=client.get_table(table),
                total_bytes_billed=query_job.total_bytes_billed,
                total_bytes_processed=query_job.total_bytes_processed,
                duration=query_job.ended - query_job.created,
            )

    raise ValueError(
        f"Invalid model materialization configuration. "
        f"Valid types are: `{', '.join((m.name for m in MaterializationTypes))}`. "
        f"Got: `{materialization}`"
    )



import dataclasses
import importlib
import inspect
import re
from collections import defaultdict
from datetime import datetime
from enum import Enum, auto
from inspect import getfile
from pathlib import Path
from types import ModuleType
from typing import (
    Any,
    Dict,
    Iterable,
    List,
    NamedTuple,
    Optional,
    Set,
    Tuple,
    Type,
    Union,
)

from pydantic import NameEmail
from sqlalchemy import Column, MetaData, Table
from sqlalchemy.orm import declared_attr, registry
from sqlalchemy.sql import ColumnCollection

from amora.config import settings
from amora.logger import logger
from amora.protocols import Compilable, CompilableProtocol
from amora.utils import ensure_path, list_files, model_path_for_target_path

metadata = MetaData(schema=f"{settings.TARGET_PROJECT}.{settings.TARGET_SCHEMA}")
mapper_registry = registry(metadata=metadata)

Model = Type["AmoraModel"]
Models = Iterable[Model]

LabelKey = str
LabelValue = str
LabelKeys = Iterable[LabelKey]

SQLALCHEMY_METADATA_KEY = "sa"


def Field(*args, **kwargs):
    return dataclasses.field(
        init=False, metadata={SQLALCHEMY_METADATA_KEY: Column(*args, **kwargs)}
    )


class Label(NamedTuple):
    """
    A label is a (key, value) pair that can also be represented as a "key:value" string. E.g.:

    ```python
    Label("freshness", "daily")
    ```
    """

    key: LabelKey
    value: LabelValue

    def __str__(self):
        return f"{self.key}:{self.value}"

    def __repr__(self):
        return str(self)

    def __eq__(self, other):
        if not isinstance(other, str):
            return self.key == other.key and self.value == other.value

        return self == Label.from_str(other)

    @classmethod
    def from_str(cls, label: str):
        """
        >>> Label.from_str("domain:health")
        Label(key="domain", value="health")
        """

        return cls(*label.split(":"))


LabelRepr = Union[Label, str]
Labels = Set[Label]


@dataclasses.dataclass
class PartitionConfig:
    field: str
    data_type: str = "date"
    granularity: str = "day"
    range: Dict[str, Any] = dataclasses.field(default_factory=dict)


class AutoName(Enum):
    def _generate_next_value_(name, start, count, last_values):
        return name


class MaterializationTypes(AutoName):
    ephemeral = auto()
    view = auto()
    table = auto()


Owner = NameEmail


@dataclasses.dataclass
class ModelConfig:
    """
    Model configuration metadata

    Attributes:
        cluster_by (List[str]): BigQuery tables can be [clustered](https://cloud.google.com/bigquery/docs/clustered-tables) to colocate related data. Expects a list of columns, as strings.
        description (Optional[str]): A string description of the model, used for documentation
        labels (Labels): Labels that can be used for data catalog and resource selection
        materialized (amora.models.MaterializationTypes): The materialization configuration: `view`, `table`, `ephemeral`. Default: `view`
        partition_by (Optional[PartitionConfig]): BigQuery supports the use of a [partition by](https://cloud.google.com/bigquery/docs/partitioned-tables) clause to easily partition a table by a column or expression. This option can help decrease latency and cost when querying large tables.
        hours_to_expire (Optional[int]): Bigquery supports setting an expiration time for a table. (https://cloud.google.com/bigquery/docs/managing-tables?hl=pt-br#updating_a_tables_expiration_time). Having this option can help us free up more space.
    """

    description: str = "Undocumented! Generated by Amora Data Build Tool 💚"
    materialized: MaterializationTypes = MaterializationTypes.view
    partition_by: Optional[PartitionConfig] = None
    cluster_by: Optional[List[str]] = None
    labels: Labels = dataclasses.field(default_factory=set)
    owner: Optional[Owner] = None
    hours_to_expire: Optional[int] = None

    @property
    def labels_dict(self) -> Dict[str, str]:
        return {label.key: label.value for label in self.labels}


class AmoraModel:
    """
    Attributes:
        __depends_on__ (Models): A list of Amora Models that the current model depends on
        __model_config__ (ModelConfig): Model configuration metadata
        __table__ (Table): SQLAlchemy table object
        __table_args__ (Dict[str, Any]): SQLAlchemy table arguments
        __tablename__override__ (Optional[str]): Set the desired table name. Overrides __tablename__
    """

    __table__: Table
    __tablename__override__: Optional[str] = None
    __depends_on__: Models = []
    __model_config__ = ModelConfig(materialized=MaterializationTypes.view)

    __sa_dataclass_metadata_key__ = SQLALCHEMY_METADATA_KEY
    __table_args__: Dict[str, Any] = {"extend_existing": True}

    def __init_subclass__(cls, **kwargs) -> None:
        super().__init_subclass__(**kwargs)
        mapper_registry.mapped(dataclasses.dataclass(cls))

    @declared_attr
    def __tablename__(cls: Model) -> str:  # type: ignore
        """
        By default, `__tablename__` is the `snake_case` class name.

        ```python
        class MyModel(AmoraModel):
            ...


        assert MyModel.__tablename__ == "my_model"
        ```
        """

        return (
            cls.__tablename__override__
            or re.sub(r"(?<!^)(?=[A-Z])", "_", cls.__mro__[0].__name__).lower()
        )

    @classmethod
    def columns(cls) -> Optional[ColumnCollection]:
        if cls.__model_config__.materialized == MaterializationTypes.ephemeral:
            cte = cls.source()
            if cte is not None:
                return cte.exported_columns
            else:
                return None
        else:
            return cls.__table__.columns

    @classmethod
    def dependencies(cls) -> List[Table]:
        source = cls.source()
        if source is None:
            return []

        return [other.__table__ for other in cls.__depends_on__]

    @classmethod
    def source(cls) -> Optional[Compilable]:
        """
        Called when `amora compile` is executed, Amora will build this model
        in your data warehouse by wrapping it in a `create view as` or `create table as` statement.

        Return `None` for defining models for tables/views that already exist on the data warehouse
        and shouldn't be managed by Amora.

        Return a `Compilable`, which is a sqlalchemy select statement, in order to compile the model with the given statement
        :return:
        """
        return None

    @classmethod
    def target_path(cls) -> Path:
        # {settings.dbt_models_path}/a_model/a_model.py -> a_model/a_model.py
        strip_path = settings.models_path.as_posix()
        relative_model_path = str(cls.path()).split(strip_path)[1][1:]
        # a_model/a_model.py -> ~/project/amora/target/a_model/a_model.sql
        target_file_path = settings.target_path.joinpath(
            relative_model_path.replace(".py", ".sql")
        )

        return target_file_path

    @classmethod
    def path(cls) -> Path:
        return Path(getfile(cls))

    @classmethod
    def unique_name(cls) -> str:
        return cls.__table__.fullname

    @classmethod
    def owner(cls) -> str:
        if model_owner := cls.__model_config__.owner:
            return str(model_owner)

        return ""


def _is_amora_model(candidate: ModuleType) -> bool:
    return (
        isinstance(candidate, CompilableProtocol)
        and inspect.isclass(candidate)
        and issubclass(candidate, AmoraModel)
        and hasattr(candidate, "__table__")
    )


@ensure_path
def amora_model_for_path(path: Path) -> Model:
    try:
        relative_module_name = (
            path.relative_to(settings.models_path)
            .as_posix()
            .replace("/", ".")
            .replace(".py", "")
        )

        module = importlib.import_module(
            relative_module_name, settings.models_path.name
        )
    except ModuleNotFoundError as e:
        raise ValueError(f"Invalid path `{path}`") from e

    compilables = inspect.getmembers(
        module,
        _is_amora_model,
    )

    for _name, class_ in compilables:
        if class_.path() == path:
            return class_

    raise ValueError(f"Invalid path `{path}`")


def amora_model_for_target_path(path: Path) -> Model:
    model_path = model_path_for_target_path(path)
    return amora_model_for_path(model_path)


def amora_model_for_name(model_name: str) -> Model:
    for m in mapper_registry.mappers:
        if m.class_.unique_name() == model_name:
            return m.class_

    raise ValueError(f"{model_name} not found on models list")


def amora_model_from_name_list(
    model_name_list: Iterable[str],
) -> Iterable[Tuple[Model, Path]]:
    for model, _path in list_models():
        if model.unique_name() in model_name_list:
            yield model, _path


def list_models(
    path: Path = settings.models_path,
) -> Iterable[Tuple[Model, Path]]:
    for model_file_path in list_files(path, suffix=".py"):
        if model_file_path.stem.startswith("_"):
            continue

        try:
            yield amora_model_for_path(model_file_path), model_file_path

        except ValueError:
            logger.exception(
                "Unable to load amora model for path",
                extra={"model_file_path": model_file_path},
            )
            continue


def list_models_with_owner(owner: Union[str, Owner]) -> Iterable[Tuple[Model, Path]]:
    if isinstance(owner, str):
        owner = Owner.validate(owner)

    for model, file_path in list_models():
        if owner == model.__model_config__.owner:
            yield model, file_path


def owners_to_models_dict() -> Dict[str, List[Model]]:
    owners_dict = defaultdict(list)
    for model, _ in list_models():
        owner = model.owner()
        if owner:
            owners_dict[owner].append(model)
    return owners_dict


def select_models_with_labels(labels: Labels) -> Iterable[Tuple[Model, Path]]:
    def matches_labels(item: Tuple[Model, Path]) -> bool:
        model, _model_path = item
        return match_labels(model, labels)

    return filter(matches_labels, list_models())


def select_models_with_label_keys(
    label_keys: LabelKeys,
) -> Iterable[Tuple[Model, Path]]:
    keys = set(label_keys)

    def matches_labels(item: Tuple[Model, Path]) -> bool:
        model, _model_path = item
        return match_label_keys(model, keys)

    return filter(matches_labels, list_models())


def match_label_keys(model: Model, label_keys: Iterable[LabelKey]) -> bool:
    for key in label_keys:
        for label in model.__model_config__.labels:
            if label.key == key:
                return True
    return False


def match_labels(model: Model, labels: Labels) -> bool:
    for label in labels:
        if label in model.__model_config__.labels:
            return True
    return False



from typing import Dict, Iterable, List, Tuple

from feast import Entity, FeatureService, FeatureView
from feast.diff.property_diff import TransitionType
from feast.diff.registry_diff import RegistryDiff
from feast.repo_contents import RepoContents
from pydantic import BaseModel
from sqlalchemy.orm import InstrumentedAttribute

from amora.feature_store.feature_view import name_for_model
from amora.feature_store.type_mapping import VALUE_TYPE_TO_NAME, value_type_for_column
from amora.models import Model, list_models

FEATURE_REGISTRY: Dict[str, Tuple[FeatureView, FeatureService, Model]] = {}


def get_entities() -> Iterable[Entity]:
    for fv, _service, model in FEATURE_REGISTRY.values():
        for entity_name in fv.entities:
            entity_column: InstrumentedAttribute = getattr(model, entity_name)

            yield Entity(
                name=entity_name,
                value_type=value_type_for_column(entity_column),
                description=entity_column.comment or "",
            )


def get_feature_views() -> List[FeatureView]:
    return [fv for (fv, _service, _model) in FEATURE_REGISTRY.values()]


def get_feature_service(model: Model) -> FeatureService:
    (_fv, service, _model) = FEATURE_REGISTRY[name_for_model(model)]
    return service


def get_feature_services() -> List[FeatureService]:
    return [service for (_fv, service, _model) in FEATURE_REGISTRY.values()]


def get_repo_contents() -> RepoContents:
    # fixme: making sure that we've collected all Feature Views
    _models = list(list_models())

    feature_views = list(set(get_feature_views()))
    entities = list(set(get_entities()))
    feature_services = list(set(get_feature_services()))
    data_sources = [fv.batch_source for fv in feature_views]

    return RepoContents(
        data_sources=data_sources,
        feature_views=feature_views,
        entities=entities,
        feature_services=feature_services,
        on_demand_feature_views=[],
        request_feature_views=[],
        stream_feature_views=[],
    )


class ObjectDiff(BaseModel):
    name: str
    transition_type: str
    object_type: str


class PropertyDiff(BaseModel):
    object_name: str
    property_name: str
    declared: str
    existing: str


class FeatureDiff(BaseModel):
    name: str
    diff: Iterable


class Diff(BaseModel):
    objects: List[ObjectDiff]
    properties: List[PropertyDiff]
    features: List[FeatureDiff]


def parse_feature_schema_diff(existing: str, declared: str):
    def parse_diff_as_feature_records(value):
        records = set()
        for item in value:
            if hasattr(item, "feature_columns"):
                return parse_diff_as_feature_records(item.feature_columns)

            records.add((item.name, VALUE_TYPE_TO_NAME[item.value_type]))

        return records

    existing_features, declared_features = parse_diff_as_feature_records(
        existing
    ), parse_diff_as_feature_records(declared)

    removed = existing_features - declared_features
    added = declared_features - existing_features

    for name, value_type in added:
        yield {"name": name, "value_type": value_type, "change": "added"}

    for name, value_type in removed:
        yield {"name": f"~~{name}~~", "value_type": value_type, "change": "removed"}


def parse_diff(registry_diff: RegistryDiff):
    objects = []
    properties = []
    features = []

    for object_diff in registry_diff.feast_object_diffs:
        if object_diff.transition_type is TransitionType.UNCHANGED:
            continue

        objects.append(
            ObjectDiff(
                name=object_diff.name,
                transition_type=object_diff.transition_type.name,
                object_type=object_diff.feast_object_type.name,
            )
        )

        for property_diff in object_diff.feast_object_property_diffs:
            if property_diff.property_name == "features":
                features.append(
                    FeatureDiff(
                        name=object_diff.name,
                        diff=parse_feature_schema_diff(
                            property_diff.val_existing,
                            property_diff.val_declared,
                        ),
                    )
                )

            else:
                properties.append(
                    PropertyDiff(
                        object_name=object_diff.name,
                        property_name=property_diff.property_name,
                        declared=str(property_diff.val_declared),
                        existing=str(property_diff.val_existing),
                    )
                )

    return Diff(
        objects=objects,
        properties=properties,
        features=features,
    )



from timeit import default_timer
from typing import Optional

from dash import Dash
from flask import Response, g, request
from prometheus_client import CollectorRegistry, Histogram
from prometheus_client.utils import INF
from prometheus_flask_exporter import PrometheusMetrics
from prometheus_flask_exporter.multiprocess import GunicornPrometheusMetrics

from amora.dash.config import settings
from amora.logger import logger
from amora.version import VERSION


def add_prometheus_metrics(dash: Dash) -> None:
    flask_app = dash.server
    registry = CollectorRegistry()
    if settings.DEBUG:
        metrics = PrometheusMetrics(
            app=flask_app, registry=registry, export_defaults=False
        )
    else:
        metrics = GunicornPrometheusMetrics(
            app=flask_app, registry=registry, export_defaults=False
        )

    metrics.info("amora_version", "Amora version", version=VERSION)

    pathname_request_duration_metric = Histogram(
        name="amora_dash_page_pathname_change_duration",
        documentation="HTTP request duration, in seconds, related to an UI URL pathname change.",
        labelnames=("method", "status"),
        unit="seconds",
        registry=metrics.registry,
    )

    component_update_request_duration_metric = Histogram(
        name="amora_dash_component_update_duration",
        documentation="HTTP request duration, in seconds, related to an UI component update.",
        labelnames=("method", "status"),
        unit="seconds",
        registry=metrics.registry,
    )

    component_update_response_size_metric = Histogram(
        name="amora_dash_component_update_response_size",
        documentation="HTTP response size, in bytes, related to an UI component update.",
        labelnames=("method", "status"),
        buckets=[*settings.METRICS_COMPONENT_UPDATE_RESPONSE_SIZE_BUCKETS, INF],
        unit="bytes",
        registry=metrics.registry,
    )

    def before_request():
        g.metrics_start_time = default_timer()

    def after_request(response: Response) -> Response:
        start_time: Optional[float] = g.pop("metrics_start_time", None)
        if not start_time:
            return response

        total_time = max(default_timer() - start_time, 0)

        if request.path != "/_dash-update-component":
            return response

        payload: dict = request.get_json()  # type: ignore
        if "_pages_location.pathname" in payload["changedPropIds"]:
            for i in payload["inputs"]:
                if i["id"] == "_pages_location" and i["property"] == "pathname":
                    pathname_request_duration_metric.labels(
                        method=request.method,
                        status=response.status_code,
                    ).observe(total_time)
                    logger.info(
                        "UI page location change", extra=dict(urlpath=i["value"])
                    )
                    return response

        elif "url.pathname" in payload["changedPropIds"]:
            return response

        else:
            inputs = ":".join(i["id"] for i in payload["inputs"])
            logger.info(
                "Component update request",
                extra=dict(
                    inputs=inputs,
                    output=payload["output"],
                    response_size=response.content_length,
                    duration=total_time,
                ),
            )
            component_update_request_duration_metric.labels(
                method=request.method,
                status=response.status_code,
            ).observe(total_time)
            component_update_response_size_metric.labels(
                method=request.method,
                status=response.status_code,
            ).observe(response.content_length)

            return response
        return response

    flask_app.before_request(before_request)
    flask_app.after_request(after_request)



import typer

app = typer.Typer(help="Amora Web UI")


@app.command("serve")
def serve():
    """
    Runs Amora's Web UI on the configured host and port.
    Debug mode can be activated with the envvar `AMORA_DASH_DEBUG=1`,
    which runs the webserver (Flask) on development mode. E.g:

    ```
    $ AMORA_DASH_DEBUG=1 amora dash serve

    Dash is running on http://127.0.0.1:8050/
     * Serving Flask app 'amora.dash.app' (lazy loading)
     * Environment: production
       WARNING: This is a development server. Do not use it in a production deployment.
       Use a production WSGI server instead.
     * Debug mode: on
    ```

    The default behaviour is a production ready setup, which runs amora on a multiprocess
    environment using [gunicorn](https://github.com/benoitc/gunicorn)

    ```
    [2022-11-09 09:31:35 -0300] [22863] [INFO] Starting gunicorn 20.1.0
    [2022-11-09 09:31:35 -0300] [22863] [INFO] Listening at: http://127.0.0.1:8050 (22863)
    [2022-11-09 09:31:35 -0300] [22863] [INFO] Using worker: sync
    [2022-11-09 09:31:35 -0300] [22875] [INFO] Booting worker with pid: 22875
    [2022-11-09 09:31:35 -0300] [22876] [INFO] Booting worker with pid: 22876
    ```

    The amount of [worker processes](https://docs.gunicorn.org/en/stable/settings.html#workers)
    can be configured with the envvar `AMORA_DASH_GUNICORN_WORKERS` (default: 2) to match
    the host hardware. The [timeout](https://docs.gunicorn.org/en/stable/settings.html#timeout)
    can be configured with the envvar `AMORA_DASH_GUNICORN_WORKER_TIMEOUT` (default: 30)

    ## Prometheus Metrics

    On the production ready setup, the metrics resource is exposed on a different port,
    by default on `:9090/metrics`, and configurable with `AMORA_DASH_METRICS_PORT`.
    On development/debug, `/metrics` share the same port as the application.

    Metrics can be disabled by setting the `AMORA_DASH_METRICS_ENABLED` to `0` or `False`.

    ```
    # HELP amora_version Multiprocess metric
    # TYPE amora_version gauge
    amora_version{version="0.1.16"} 1.0
    # HELP amora_dash_component_update_duration_seconds Multiprocess metric
    # TYPE amora_dash_component_update_duration_seconds histogram
    amora_dash_component_update_duration_seconds_sum{method="POST",status="200"} 0.030016314999699034
    amora_dash_component_update_duration_seconds_bucket{le="0.005",method="POST",status="200"} 14.0
    amora_dash_component_update_duration_seconds_bucket{le="0.01",method="POST",status="200"} 16.0
    amora_dash_component_update_duration_seconds_bucket{le="0.025",method="POST",status="200"} 16.0
    amora_dash_component_update_duration_seconds_bucket{le="0.05",method="POST",status="200"} 16.0
    amora_dash_component_update_duration_seconds_bucket{le="0.075",method="POST",status="200"} 16.0
    amora_dash_component_update_duration_seconds_bucket{le="0.1",method="POST",status="200"} 16.0
    amora_dash_component_update_duration_seconds_bucket{le="0.25",method="POST",status="200"} 16.0
    amora_dash_component_update_duration_seconds_bucket{le="0.5",method="POST",status="200"} 16.0
    amora_dash_component_update_duration_seconds_bucket{le="0.75",method="POST",status="200"} 16.0
    amora_dash_component_update_duration_seconds_bucket{le="1.0",method="POST",status="200"} 16.0
    amora_dash_component_update_duration_seconds_bucket{le="2.5",method="POST",status="200"} 16.0
    amora_dash_component_update_duration_seconds_bucket{le="5.0",method="POST",status="200"} 16.0
    amora_dash_component_update_duration_seconds_bucket{le="7.5",method="POST",status="200"} 16.0
    amora_dash_component_update_duration_seconds_bucket{le="10.0",method="POST",status="200"} 16.0
    amora_dash_component_update_duration_seconds_bucket{le="+Inf",method="POST",status="200"} 16.0
    amora_dash_component_update_duration_seconds_count{method="POST",status="200"} 16.0
    ...
    ```

    ## User Authentication

    ![amora dash authentication](../static/user-guide/web-ui/web-ui-auth.gif)

    User authentication on the interface can be achieved with [auth0](http://auth0.com)
    by running `amora dash serve` with the following envvars:

    ```
    AMORA_DASH_AUTH0_CLIENT_ID=YOUR_CLIENT_ID
    AMORA_DASH_AUTH0_CLIENT_SECRET=YOUR_CLIENT_SECRET
    AMORA_DASH_AUTH0_DOMAIN=YOUR_DOMAIN
    ```

    ## Theming

    The interface style can be easily replaced by setting the envvar `AMORA_DASH_DBC_THEME`.

    `AMORA_DASH_DBT_THEME=SKETCHY` (Default):

    ![sketchy theme](../static/user-guide/web-ui/web-ui-themes-sketchy.png)

    `AMORA_DASH_DBT_THEME=MINTY`

    ![minty theme](../static/user-guide/web-ui/web-ui-themes-minty.png)

    `AMORA_DASH_DBT_THEME=CYBORG`

    ![cyborg theme](../static/user-guide/web-ui/web-ui-themes-cyborg.png)

    `AMORA_DASH_DBT_THEME=VAPOR`

    ![vapor theme](../static/user-guide/web-ui/web-ui-themes-vapor.png)

    """
    from amora.dash.app import dash_app
    from amora.dash.config import settings
    from amora.dash.gunicorn.application import StandaloneApplication
    from amora.dash.gunicorn.config import child_exit, when_ready

    if settings.DEBUG:
        return dash_app.run(
            debug=settings.DEBUG, host=settings.HTTP_HOST, port=settings.HTTP_PORT
        )

    options = {
        "bind": f"{settings.HTTP_HOST}:{settings.HTTP_PORT}",
        "workers": settings.GUNICORN_WORKERS,
        "timeout": settings.GUNICORN_WORKER_TIMEOUT,
    }
    if settings.METRICS_ENABLED:
        options.update(
            {
                "when_ready": when_ready,
                "child_exit": child_exit,
            }
        )

    StandaloneApplication(app=dash_app.server, options=options).run()


@app.command(
    "inspect",
    help="Inspect the project data queries and generates a cache to speedup the interface",
)
def inspect():
    from concurrent.futures import ThreadPoolExecutor

    from amora.config import settings as amora_settings
    from amora.dash.config import settings
    from amora.logger import logger
    from amora.meta_queries import summarize
    from amora.models import list_models
    from amora.providers.bigquery import sample
    from amora.questions import QUESTIONS

    if not amora_settings.STORAGE_CACHE_ENABLED:
        logger.debug("Cache disabled. Skipping cache generation.")
        return

    with ThreadPoolExecutor(
        max_workers=settings.THREAD_POOL_EXECUTOR_WORKERS
    ) as executor:
        # cache model summary
        for model, path_ in list_models():
            executor.submit(summarize, model)

        # cache model data sample
        for model, path_ in list_models():
            executor.submit(sample, model)

        # cache data questions
        for question in QUESTIONS:
            executor.submit(question.answer_df)



from typing import Any, Callable, List, Optional

import dash_bootstrap_components as dbc
import pandas as pd
import plotly
import plotly.express as px
from dash import dcc, html
from dash.development.base_component import Component
from pandas.api.types import (
    is_array_like,
    is_bool_dtype,
    is_datetime64_any_dtype,
    is_numeric_dtype,
    is_string_dtype,
)
from plotly.graph_objects import Figure

from amora.models import Model
from amora.providers.bigquery import sample


def figure_apply_default_layout(figure: Figure) -> Figure:
    figure.update_layout(
        showlegend=False,
        bargap=0,
        margin=dict(l=0, r=0, b=0, t=0),
        yaxis_title=None,
        xaxis_title=None,
        width=200,
        height=200,
    )

    figure.update_yaxes(visible=False, showticklabels=False)

    return figure


def value_counts_series_to_dataframe(series: pd.Series) -> pd.DataFrame:
    return series.to_frame().reset_index().set_axis(["x", "count"], axis=1)


def cut_formatter(cut: pd.Series, formatter_function: Callable) -> List[str]:
    """
    Apply a formatter function in both `left` and `right` Interval.
    Sometimes the pd.cuts generates long Interval strings, and we may
    want to format each part of it.
    """
    return [
        f"{formatter_function(interval.left)}-{formatter_function(interval.right)}"
        for _, interval in cut.iteritems()
    ]


def get_binned_dataframe(
    series: pd.Series,
    formatter_function: Optional[Callable] = None,
) -> pd.DataFrame:
    bins = min(series.nunique(), 10)
    cut = pd.cut(x=series, bins=bins, include_lowest=True, precision=0, right=False)

    if formatter_function:
        formatted = cut_formatter(cut, formatter_function)
        frequencies = pd.value_counts(formatted, sort=False)
    else:
        frequencies = pd.value_counts(cut, sort=False)

    df_frequencies = (
        frequencies.to_frame()
        .reset_index()
        .set_axis(["x", "count"], axis=1)
        .sort_values(by="x")
    )

    df_frequencies["x"] = df_frequencies["x"].astype(str)

    return df_frequencies


def datetime_series_aggregation(series: pd.Series) -> pd.Series:
    days_diff = (series.max() - series.min()).days

    if days_diff > 365:
        agg_type = "Y"
    elif days_diff > 31:
        agg_type = "M"
    else:
        agg_type = "D"

    return series.dt.to_period(agg_type).astype(str)


def create_component_for_one_unique_value(series: pd.Series) -> html.P:
    return html.P(f"One unique value: {series[0]}", style={"color": "red"})


def get_most_common_values(
    df_value_counts_normalized: pd.DataFrame,
) -> List[Component]:
    elements = []

    for _, row in df_value_counts_normalized.head(2).iterrows():
        percentage = round((row["count"] * 100), 2)
        elements.append(
            dbc.Row(
                [
                    dbc.Col(
                        f"{row['x']}",
                        style={
                            "text-overflow": "ellipsis",
                            "overflow": "hidden",
                            "white-space": "nowrap",
                            "max-width": "200px",
                        },
                    ),
                    dbc.Col(html.B(f"{percentage}%")),
                ]
            )
        )

    if df_value_counts_normalized.shape[0] > 2:
        other = round(
            (
                1
                - (
                    df_value_counts_normalized["count"][0]
                    + df_value_counts_normalized["count"][1]
                )
            )
            * 100,
            2,
        )
        elements.append(dbc.Row([dbc.Col("other"), dbc.Col(html.B(f"{other}%"))]))

    return elements


def create_bar_plot(
    df_value_counts: pd.DataFrame,
    **bar_kwargs: Any,
) -> dcc.Graph:
    figure = figure_apply_default_layout(
        px.bar(
            df_value_counts,
            x="x",
            y="count",
            **bar_kwargs,
        )
    )

    return dcc.Graph(figure=figure)


def create_component_bool_type(series: pd.Series) -> dcc.Graph:
    series_value_counts = series.value_counts()
    series_value_counts.index = series_value_counts.index.map(str)
    df_value_counts = value_counts_series_to_dataframe(series_value_counts)
    return create_bar_plot(
        df_value_counts,
    )


def create_component_numeric_type(series: pd.Series) -> dcc.Graph:
    if series.nunique() <= 2:
        return create_component_bool_type(series)
    else:
        figure = plotly.hist_series(series)
        return dcc.Graph(figure=figure_apply_default_layout(figure))


def create_component_datetime_type(series: pd.Series) -> dcc.Graph:
    datetime_series_agg = datetime_series_aggregation(series)
    series_value_counts = datetime_series_agg.value_counts()
    df_value_counts = value_counts_series_to_dataframe(series_value_counts)
    return create_bar_plot(
        df_value_counts,
    )


def create_component_string_type(series: pd.Series) -> dcc.Graph:
    series_value_counts = series.value_counts(normalize=True)
    df_value_counts_normalized = value_counts_series_to_dataframe(
        series_value_counts,
    )
    return get_most_common_values(df_value_counts_normalized)


def visualization_for_series(series: pd.Series) -> dbc.Card:
    if is_array_like(series[0]):
        component_title = f"⚠️ {series.name}"
        component_viz = dbc.Alert(
            "Visualization not implemented for arrays", color="info"
        )
    elif series.nunique() == 1:
        component_title = str(series.name)
        component_viz = create_component_for_one_unique_value(series)
    elif is_bool_dtype(series):
        component_title = f"✔️ {series.name}"
        component_viz = create_component_bool_type(series)

    elif is_numeric_dtype(series):
        component_title = f"🔢 {series.name}"
        component_viz = create_component_numeric_type(series)

    elif is_datetime64_any_dtype(series):
        component_title = f"📅 {series.name}"
        component_viz = create_component_datetime_type(series)

    elif is_string_dtype(series):
        component_title = f"🔤 {series.name}"
        component_viz = create_component_string_type(series)
    else:
        component_title = f"❓ {series.name}"
        component_viz = html.P("No viz implemented for this type of column 😔")

    return dbc.Card(
        [dbc.CardHeader(component_title), dbc.CardBody(component_viz)],
        className="w-75 mb-3",
    )


def create_df_visualizations(df: pd.DataFrame) -> Component:
    col_list = [
        dbc.Col(
            visualization_for_series(series=df[column_name]),
            width=4,
        )
        for column_name in df
    ]

    return dbc.Row(col_list)


def component(model: Model) -> Component:
    try:
        df = sample(model, percentage=1)
    except ValueError:
        return html.Div(
            f"This component depends on `amora.providers.bigquery.sample`, which isn't available for model `{model.unique_name}`"
        )
    else:
        return dbc.Row(
            [
                dbc.Row(
                    dbc.Alert(
                        "The visualizations bellow are generated from a 1% model sample data",
                        color="info",
                    )
                ),
                dbc.Row(create_df_visualizations(df)),
            ]
        )



import numpy
import pandas as pd
from dash import dash_table, html

from amora.models import Model
from amora.providers.bigquery import sample


def _adapt_dataframe_values(df: pd.DataFrame) -> pd.DataFrame:
    """
    Transforma object em strings, porque dash_table.DataTable suporta apenas os tipos string, number e boolean
    """
    for col in df.columns:
        if df[col].dtype == numpy.dtype("O"):
            df[col] = df[col].astype(str)

    return df


def component(model: Model) -> dash_table.DataTable:
    try:
        df = sample(model, percentage=1)
    except ValueError:
        return html.Div(f"Sample not implemented for model {model.unique_name()}")

    df = _adapt_dataframe_values(df)

    return dash_table.DataTable(
        data=df.to_dict("records"),
        columns=[{"name": col, "id": col} for col in sorted(df.columns.values)],
        export_format="csv",
        sort_action="native",
    )



from feast import ValueType
from feast.types import FeastType, from_value_type
from sqlalchemy import Column
from sqlalchemy.sql import sqltypes

_SQLALCHEMY_TYPES_TO_FS_TYPES = {
    sqltypes.Float: ValueType.FLOAT,
    sqltypes.String: ValueType.STRING,
    sqltypes.Integer: ValueType.INT64,
    sqltypes.Boolean: ValueType.BOOL,
    sqltypes.TIMESTAMP: ValueType.UNIX_TIMESTAMP,
    sqltypes.NullType: ValueType.NULL,
    (sqltypes.ARRAY, sqltypes.String): ValueType.STRING_LIST,
    (sqltypes.ARRAY, sqltypes.Integer): ValueType.INT64_LIST,
    (sqltypes.ARRAY, sqltypes.Boolean): ValueType.BOOL_LIST,
    (sqltypes.ARRAY, sqltypes.Float): ValueType.FLOAT_LIST,
}

VALUE_TYPE_TO_NAME = {t.value: t.name for t in ValueType}


def value_type_for_column(col: Column) -> ValueType:
    if isinstance(col.type, sqltypes.ARRAY):
        key = (col.type.__class__, col.type.item_type.__class__)
    else:
        key = col.type.__class__

    return _SQLALCHEMY_TYPES_TO_FS_TYPES[key]


def feast_type_for_colum(col: Column) -> FeastType:
    return from_value_type(value_type_for_column(col))



from concurrent import futures
from typing import Dict, List, Optional

import pytest
import typer

from amora import compilation, manifest, materialization, utils
from amora.cli import dash, feature_store, models
from amora.cli.shared_options import force_option, models_option, target_option
from amora.cli.type_specs import Models
from amora.config import settings
from amora.dag import DependencyDAG
from amora.models import list_models

app = typer.Typer(
    pretty_exceptions_enable=False,
    help="Amora Data Build Tool enables engineers to transform data in their warehouses "
    "by defining schemas and writing select statements with SQLAlchemy. Amora handles turning these "
    "select statements into tables and views",
)


@app.command()
def compile(
    models: Optional[Models] = models_option,
    target: Optional[str] = target_option,
    force: Optional[bool] = force_option,
) -> None:
    """
    Generates executable SQL from model files. Compiled SQL files are written to the `./target` directory.
    """

    current_manifest = manifest.Manifest.from_project()
    previous_manifest = manifest.Manifest.load()

    if force or not previous_manifest:
        compilation.remove_compiled_files()
        models_to_compile = list_models()
    else:
        removed = previous_manifest.models.keys() - current_manifest.models.keys()
        compilation.remove_compiled_files(removed)
        models_to_compile = current_manifest.get_models_to_compile(previous_manifest)

    for model, model_file_path in models_to_compile:
        if models and model_file_path.stem not in models:
            continue

        source_sql_statement = model.source()
        if source_sql_statement is None:
            typer.echo(f"⏭ Skipping compilation of model `{model_file_path}`")
            continue

        target_file_path = model.target_path()
        typer.echo(f"🏗 Compiling model `{model_file_path}` -> `{target_file_path}`")

        content = compilation.compile_statement(source_sql_statement)
        target_file_path.parent.mkdir(parents=True, exist_ok=True)
        target_file_path.write_text(content)

    current_manifest.save()


@app.command()
def materialize(
    models: Optional[Models] = models_option,
    target: str = target_option,
    draw_dag: bool = typer.Option(False, "--draw-dag"),
    no_compile: bool = typer.Option(
        False,
        "--no-compile",
        help="Don't run `amora compile` before the materialization",
    ),
) -> None:
    """
    Executes the compiled SQL against the current target database.

    """
    if not no_compile:
        compile(models=models, target=target)

    model_to_task: Dict[str, materialization.Task] = {}

    for target_file_path in utils.list_target_files():
        if models and target_file_path.stem not in models:
            continue

        task = materialization.Task.for_target(target_file_path)
        model_to_task[task.model.unique_name()] = task

    dag = DependencyDAG.from_tasks(tasks=model_to_task.values())

    if draw_dag:
        dag.draw()

    with futures.ProcessPoolExecutor(
        max_workers=settings.MATERIALIZE_NUM_THREADS
    ) as executor:
        for models_to_materialize in dag.topological_generations():
            current_tasks: List[materialization.Task] = []
            for model_name in models_to_materialize:
                if model_name in model_to_task:
                    current_tasks.append(model_to_task[model_name])
                else:
                    typer.echo(f"⚠️  Skipping `{model_name}`")
                    continue

            if not current_tasks:
                continue

            results = executor.map(
                materialization.materialize,
                [current_task.sql_stmt for current_task in current_tasks],
                [current_task.model.unique_name() for current_task in current_tasks],
                [current_task.model.__model_config__ for current_task in current_tasks],
            )

            for result in results:
                if result:
                    typer.echo(result)


@app.command(
    context_settings={"allow_extra_args": True, "ignore_unknown_options": True}
)
def test(ctx: typer.Context) -> None:
    """
    Runs tests on data in deployed models. Run this after `amora materialize`
    to ensure that the data state is up-to-date. Optional arguments are passed
    to pytest.
    """

    pytest_args = settings.DEFAULT_PYTEST_ARGS + ctx.args
    return_code = pytest.main(pytest_args)

    raise typer.Exit(return_code)


app.add_typer(dash.app, name="dash")
app.add_typer(models.app, name="models")
app.add_typer(feature_store.app, name="feature-store")
